import functools
import pika
import time
import json

from varys.process import Process


class Producer(Process):
    def __init__(
        self,
        message_queue,
        exchange,
        configuration,
        log_file,
        log_level,
        queue_suffix,
        exchange_type,
        routing_key="arbitrary_string",
        reconnect_wait=10,
    ):
        super().__init__(
            message_queue,
            exchange,
            configuration,
            log_file,
            log_level,
            queue_suffix,
            exchange_type,
            routing_key=routing_key,
            reconnect_wait=reconnect_wait,
        )

        self._message_number = 0

        self._message_properties = pika.BasicProperties(
            content_type="json",
            delivery_mode=pika.DeliveryMode.Persistent,
        )

    def publish_message(self, message, max_attempts=3):
        try:
            message_str = json.dumps(message, ensure_ascii=False)
        except TypeError:
            self._log.error(f"Unable to serialise message into json: {str(message)}")

        attempt = 0
        while attempt < max_attempts:
            try:
                attempt += 1
                self._log.info(
                    f"Sending message (attempt {attempt}): {json.dumps(message)}"
                )
                self._connection.add_callback_threadsafe(
                    functools.partial(
                        self._channel.basic_publish,
                        self._exchange,
                        self._routing_key,
                        message_str,
                        self._message_properties,
                        mandatory=True,
                    )
                )
            # except pika.exceptions.ConnectionWrongStateError:
            except Exception:
                self._log.exception(
                    f"Exception while trying to publish message on attempt {attempt}!:"
                )

                if attempt < max_attempts:
                    if self._reconnect_wait > 0:
                        time.sleep(self._reconnect_wait)

                    continue
                else:
                    raise

            break

        self._message_number += 1
        self._log.info(f"Published message #{self._message_number}")

    def run(self):
        while not self._stopping:
            try:
                self._connection = pika.BlockingConnection(self._parameters)
                self._channel = self._connection.channel()
                self._channel.exchange_declare(
                    exchange=self._exchange,
                    exchange_type=self._exchange_type,
                    durable=True,
                )
                self._channel.queue_declare(queue=self._queue, durable=True)
                self._channel.queue_bind(
                    queue=self._queue,
                    exchange=self._exchange,
                    routing_key=self._routing_key,
                )
                self._channel.confirm_delivery()
                # time_limit=None leads to the connection being dropped for inactivity
                # not sure if this should be while not self._stopping
                # while true:
                while not self._stopping:
                    self._connection.process_data_events(time_limit=1)
            except Exception:
                self._log.exception("Producer caught exception:")

            if self._stopping or self._reconnect_wait < 0:
                break
            else:
                time.sleep(self._reconnect_wait)
                continue

    def stop(self):
        self._log.info("Stopping producer as instructed...")
        # probably have to say we're closing so run doesn't try to reopen connection
        self._stopping = True

        self._connection.add_callback_threadsafe(
            functools.partial(self._connection.process_data_events, time_limit=3)
        )

        self._connection.add_callback_threadsafe(self._channel.close)
        self._connection.add_callback_threadsafe(self._connection.close)

        self._log.debug("Stopping producer logger...")
        self._stop_logger()

        self._log.info("Stopped producer as instructed.")



from threading import Thread
from os import path
import ssl
import logging

import pika


class Process(Thread):
    def __init__(
        self,
        message_queue,
        exchange,
        configuration,
        log_file,
        log_level,
        queue_suffix,
        exchange_type,
        routing_key="arbitrary_string",
        reconnect_wait=10,
    ):
        super().__init__()

        self._message_queue = message_queue
        self._routing_key = routing_key
        self._exchange = exchange
        self._queue = exchange + "." + queue_suffix
        self._log_file = path.abspath(log_file)  # so we know which file handle to drop when we stop
        self._setup_logger(log_level)

        self._connection = None
        self._channel = None
        self._stopping = False
        self._reconnect_wait = reconnect_wait

        if exchange_type == "fanout":
            self._exchange_type = pika.exchange_type.ExchangeType.fanout
        elif exchange_type == "topic":
            self._exchange_type = pika.exchange_type.ExchangeType.topic
        elif exchange_type == "direct":
            self._exchange_type = pika.exchange_type.ExchangeType.direct
        elif exchange_type == "headers":
            self._exchange_type = pika.exchange_type.ExchangeType.headers
        else:
            raise ValueError(
                "Exchange type must be one of: fanout, topic, direct, headers"
            )

        if configuration.use_tls:
            context = ssl.create_default_context(
                purpose=ssl.Purpose.SERVER_AUTH,
                cafile=configuration.ca_certificate,
            )
            # default behaviour for Purpose.SERVERAUTH
            context.verify_mode = ssl.CERT_REQUIRED
            context.check_hostname = True

            context.load_cert_chain(configuration.client_certificate,
                                    configuration.client_key)

            ssl_options = pika.SSLOptions(context, configuration.ampq_url)
        else:
            ssl_options = None

        self._parameters = pika.ConnectionParameters(
            host=configuration.ampq_url,
            port=configuration.port,
            credentials=pika.PlainCredentials(
                username=configuration.username, password=configuration.password
            ),
            ssl_options=ssl_options,
        )

    def _setup_logger(self, log_level):
        name = self._exchange
        log_path = self._log_file

        self._log = logging.getLogger(name)
        self._log.propagate = False
        self._log.setLevel(log_level)

        # if the filename is already associated with a handler, increase the count
        try:
            handler_filenames = [fh.baseFilename for fh in self._log.handlers]
            index = handler_filenames.index(log_path)
            self._log.handlers[index].count += 1
        # otherwise create a new filehandler (with initial count 1)
        except ValueError:
            logging_fh = logging.FileHandler(log_path)
            logging_fh.setFormatter(
                logging.Formatter(
                    "%(name)s\t::%(levelname)s::%(asctime)s::\t%(message)s"
                )
            )
            self._log.addHandler(logging_fh)
            self._log.handlers[-1].count = 1

    def _stop_logger(self):
        log_path = self._log_file

        handler_filenames = [fh.baseFilename for fh in self._log.handlers]
        index = handler_filenames.index(log_path)
        self._log.handlers[index].count -= 1

        if self._log.handlers[index].count == 0:
            handler = self._log.handlers.pop(index)
            handler.close()



from collections import namedtuple
import sys
import json
import os


varys_message = namedtuple("varys_message", "basic_deliver properties body")


class configurator:
    def __init__(self, profile, config_path):
        try:
            with open(config_path, "rt") as config_fh:
                config_obj = json.load(config_fh)
        except Exception as e:
            print(
                f"Varys Configuration JSON parsing failed with exception:\n{e}",
                file=sys.stderr,
            )
            sys.exit(11)

        if config_obj["version"] != "0.1":
            print(
                "Version number in the ROZ configuration file does not appear to be current, ensure configuration format is correct if you experience errors",
                file=sys.stderr,
            )

        profile_dict = config_obj["profiles"].get(profile)
        if profile_dict:
            try:
                self.profile = profile
                self.username = profile_dict["username"]
                self.password = profile_dict["password"]
                self.ampq_url = profile_dict["amqp_url"]
                self.port = int(profile_dict["port"])

                self.use_tls = profile_dict["use_tls"]

                if self.use_tls:
                    if os.environ.get("VARYS_CA_CERTIFICATE"):
                        self.ca_certificate = os.environ.get("VARYS_CA_CERTIFICATE")

                    if profile_dict.get("ca_certificate"):
                        self.ca_certificate = profile_dict.get("ca_certificate", None)

                    if os.environ.get("VARYS_CLIENT_CERTIFICATE"):
                        self.client_certificate = os.environ.get("VARYS_CLIENT_CERTIFICATE")

                    if profile_dict.get("client_certificate"):
                        self.client_certificate = profile_dict["client_certificate"]

                    if os.environ.get("VARYS_CLIENT_KEY"):
                        self.client_key = os.environ.get("VARYS_CLIENT_KEY")

                    if profile_dict.get("client_key"):
                        self.client_key = profile_dict["client_key"]
                
                    if not self.ca_certificate or not self.client_certificate or not self.client_key:
                        print(
                            "Varys configuration JSON does not appear to contain the necessary fields for TLS configuration",
                            file=sys.stderr,
                        )
                        sys.exit(11)
                    
                    if not os.path.exists(self.ca_certificate) or not os.path.exists(self.client_certificate) or not os.path.exists(self.client_key):
                        print(
                            "Varys configuration JSON does not appear to contain the necessary files for TLS configuration",
                            file=sys.stderr,
                        )
                        sys.exit(11)

                        

            except KeyError:
                print(
                    f"Varys configuration JSON does not appear to contain the necessary fields for profile: {profile}",
                    file=sys.stderr,
                )
                sys.exit(11)
        else:
            print(
                f"Varys configuration JSON does not appear to contain the specified profile {profile}",
                file=sys.stderr,
            )
            sys.exit(2)
